from .generic_session import GenericSession
from .payload import Payload
from .pipeline import Pipeline
from .instance import Instance
from .plugin_template import CustomPluginTemplate
from .distributed_custom_code_presets import DistributedCustomCodePresets



from .logging import Logger


class BaseDecentrAIObject(object):
  """
  Generic class

  Instructions:

    1. use `super().__init__(**kwargs)` at the end of child `__init__`
    2. define `startup(self)` method for the child class and call 
       `super().startup()` at beginning of `startup()` method

      OR

    use `super().__init__(**kwargs)` at beginning of child `__init__` and then
    you can safely proceed with other initilization 

  """

  def __init__(self, log: Logger,
               create_logger=False,
               DEBUG=False,
               show_prefixes=False,
               prefix_log=None,
               log_at_startup=False,
               silent=False,
               local_cache_base_folder='.',
               local_cache_app_folder='_local_cache',
               **kwargs):

    super(BaseDecentrAIObject, self).__init__()

    if (log is None) or not hasattr(log, '_logger'):
      if not create_logger:
        raise ValueError("Logger object is invalid: {}".format(log))
      else:
        log = Logger(
          "R1", 
          DEBUG=DEBUG, 
          base_folder=local_cache_base_folder, 
          app_folder=local_cache_app_folder,
          silent=silent,
        )
    # endif

    self.log = log
    self.show_prefixes = show_prefixes
    self.prefix_log = prefix_log
    self.config_data = self.log.config_data
    self.DEBUG = DEBUG
    self.log_at_startup = log_at_startup

    if not hasattr(self, '__name__'):
      self.__name__ = self.__class__.__name__
    self.startup()

    return

  def startup(self):
    self.log.set_nice_prints()
    ver = ''
    if hasattr(self, '__version__'):
      ver = 'v.' + self.__version__
    if hasattr(self, 'version'):
      ver = 'v.' + self.version

    if self.log_at_startup:
      self.P("{}{} startup.".format(self.__class__.__name__, ' ' + ver if ver != '' else ''))
    return

  def shutdown(self):
    self.P("Shutdown in progress...")
    _VARS = ['sess', 'session']
    for var_name in _VARS:
      if vars(self).get(var_name, None) is not None:
        self.P("Warning: {} property {} still not none before closing".format(
          self.__class__.__name__, var_name), color='r')
    return

  def P(self, s, t=False, color=None, prefix=False, **kwargs):
    if self.show_prefixes or prefix:
      msg = "[{}] {}".format(self.__name__, s)
    else:
      if self.prefix_log is None:
        msg = "{}".format(s)
      else:
        msg = "{} {}".format(self.prefix_log, s)
      # endif
    # endif
    _r = self.log.P(msg, show_time=t, color=color, **kwargs)
    return _r

  def D(self, s, t=False, color=None, prefix=False, forced_debug=False, **kwargs):
    _r = -1
    if self.DEBUG or forced_debug:
      if color is None:
        color = 'd'
      if self.show_prefixes:
        msg = "[DEBUG] {}: {}".format(self.__name__, s)
      else:
        if self.prefix_log is None:
          msg = "[DEBUG] {}".format(s)
        else:
          msg = "[DEBUG]{} {}".format(self.prefix_log, s)
        # endif
      # endif
      _r = self.log.P(msg, show_time=t, color=color, prefix=prefix, **kwargs)
    # endif
    return _r

  def start_timer(self, tmr_id):
    if hasattr(self, '_timers_section'):
      return self.log.start_timer(tmr_id, section=self._timers_section)
    else:
      return self.log.start_timer(sname=self.__name__ + '_' + tmr_id)

  def end_timer(self, tmr_id, skip_first_timing=True, **kwargs):
    if hasattr(self, '_timers_section'):
      return self.log.end_timer(tmr_id, section=self._timers_section)
    else:
      return self.log.end_timer(
        sname=self.__name__ + '_' + tmr_id,
        skip_first_timing=skip_first_timing,
        **kwargs
      )

  def raise_error(self, error_text):
    """
    logs the error and raises it
    """
    self.P("{}: {}".format(self.__class__.__name__, error_text))
    raise ValueError(error_text)

  def timer_name(self, name=''):
    tn = ''
    if name == '':
      tn = self.__class__.__name__
    else:
      tn = '{}__{}'.format(self.__class__.__name__, name)
    return tn

  def __repr__(self):
    # Get the name of the class
    class_name = self.__class__.__name__

    # Get public properties (those not starting with "_")
    public_properties = {k: v for k, v in self.__dict__.items() if not k.startswith("_")}

    # Convert properties to a string representation
    properties_str = ", ".join(f"{k}={v!r}" for k, v in public_properties.items())

    return f"{class_name}({properties_str})"



# local dependencies
from ...io_formatter.base import BaseFormatter


class DefaultFormatter(BaseFormatter):

  def __init__(self, log, **kwargs):
    super(DefaultFormatter, self).__init__(
        log=log, prefix_log='[DEFAULT-FMT]', **kwargs)
    return

  def startup(self):
    pass

  def _encode_output(self, output):
    return output

  def _decode_output(self, encoded_output):
    return encoded_output

  def _decode_streams(self, dct_config_streams):
    return dct_config_streams



"""
cli_commands.py
================

This module defines the command-line interface (CLI) commands 
for the Ratio1 SDK client. It provides a structured way to interact
with the SDK's functionalities through various commands and options
and it is used by cli.py

"""

from ratio1.cli.nodes import (
  get_nodes, get_supervisors,
  restart_node, shutdown_node,
  get_apps, inspect_node
)
from ratio1.cli.oracles import get_availability, oracle_rollout
from ratio1.utils.config import (
  show_config, reset_config, show_address, get_set_network,
  get_networks, get_set_alias, get_eth_addr
)
from ratio1.cli.package_update import update_package

# Define the available commands
CLI_COMMANDS = {
    "get": {
        "nodes": {
            "func": get_nodes,
            "description": "Get the list of nodes available in the network.",
            "params": {
                ### use "(flag)" at the end of the description to indicate a boolean flag
                ### otherwise it will be treated as a str parameter
                "--all": "Get all known nodes including those that have been gone missing (flag)",  # DONE
                "--online" : "Get only online nodes as seen by a active supervisor (flag)", # DONE
                "--peered": "Get only peered nodes - ie nodes that can be used by current client address (flag)",  # DONE
                "--supervisor" : "Use a specific supervisor node",
                "--alias" : "Use a specific node alias filter",
                "--eth" : "Use a specific node (flag)",
                "--wide" : "Display all available information (flag)",
            }
        },
        "supervisors": {
            "func": get_supervisors, # DONE
            "description": "Get the list of available supervisor oracles",
            "params": {
                "--eth" : "Use a specific node (flag)",
                "--wide" : "Display all available information (flag)",
            }
        },
        "eth" : {
            "func": get_eth_addr,
            "description": "Get the ETH address given a node address",
            "params": {
                "node": "The node address to get the ETH address for"
            }
        },
        "avail": {    
            "func": get_availability,
            "description": "Get the availability of a node via the oracle network.",
            "params": {
                ### use "(flag)" at the end of the description to indicate a boolean flag
                ### otherwise it will be treated as a str parameter
                "node": "The ETH address of the node to be checked via the oracle network.",
                "--start": "The start epoch number to check the availability from",
                "--end": "The end epoch number to check the availability to",
                "--json": "Enable full JSON oracle network output (flag)",
                "--rounds": "The number of rounds to check the availability for testing purposes (default=1)",
            }
        },
        "apps": {
            "func": get_apps,
            "description": "Get the apps running on a given node, if the client is allowed on that node.",
            "params": {
                "--node": "The ETH address or the specific address of the node to get the apps from",
                "--owner" : "Get the apps for a particular owner/initiator",
                "--full": "Include admin apps (flag)",
                "--json": "Output the entire JSON config of applications (flag)",
                "--wide": "Display all available information (flag)",
            }
        },
        "networks": {
            "func": get_networks, # DONE
            "description": "Show the network configuration",
        },
    },
    "config": {
        "show": {
            "func": show_config, # DONE
            "description": "Show the current configuration including the location",
        },
        "reset": {
            "func": reset_config, # DONE
            "description": "Reset the configuration to default",
            # "params": {
            #   ### use "(flag)" at the end of the description to indicate a boolean flag 
            #   ### otherwise it will be treated as a str parameter
            #   "--force": "Force reset (flag)",  # DONE
            # }
        },
        "addr": {
            "func": show_address, # DONE
            "description": "Show the current client address",
        },
        
        "network": {
            "func": get_set_network, # DONE
            "description": "Get/Set network",
            "params": {
                "--new": "The network to set either 'mainnet' or 'testnet' (same as --set)",
                "--set": "The network to set either 'mainnet' or 'testnet' (same as --new)",
            }
        },
        
        "alias": {
            "func": get_set_alias, # DONE
            "description": "Show and sets the current client alias",
            "params": {
              "--set": "The alias to set for this SDK client",
            }
        }        
    },
    "restart": {
        "func": restart_node,
        "description": "Restart a node",
        "params": {
            "node": "The node to restart",
            "--ignore-peering": "Ignore peering when running the command (flag)",
        }
    },
    "shutdown": {
        "func": shutdown_node,
        "description": "Shutdown a node",
        "params": {
            "node": "The node to shutdown",
            "--ignore-peering": "Ignore peering when running the command (flag)",
        }
    },
    "update": {
        "func": update_package,
        "description": "Update the Ratio1 SDK client package",
        "params": {
            "--quiet": "Run the update in quiet mode (flag)",
        }
    },
    "oracle-rollout": {
        "func": oracle_rollout,
        "description": "Rollout update on all nodes in the network. The rollout order is seed nodes -> oracle nodes -> all other edge nodes. This command is needed when defining new environment variables in seed nodes, in order to make it available to all nodes in the network.",
        "params": {
            "--skip-seeds": "Skip the seed nodes in the rollout (flag)",
            "--skip-oracles": "Skip the oracle nodes in the rollout (flag)",
            "--no-timeout": "Do not wait between nodes restarts (flag)",
        }
    },
    "inspect": {
        "func": inspect_node,
        "description": "Inspect a node by address or alias.",
        "params": {
            "node": "The node address or alias to inspect",
            "--wide": "Display all available information (flag)"
        }
    },
}



import pandas as pd
from ratio1 import Session


if __name__ == '__main__':
  sess = Session(silent=True, verbosity=3)
  sess.P(sess.get_client_address(), color='g')
  
  node = '0xai_AxLOEgr3I1SCi3wp1c3tYxgVEpZrfV_qpDoG3_J8Sc4e'
  df = sess.get_nodes_apps(node=node, show_full=True, as_json=False, as_df=True)
  sess.P(f"Results:\n{df}", color='b', show=True)
    
  sess.wait(seconds=15, close_session=True)


import base64
import hashlib
import os
import binascii
import zlib
import json

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.kdf.hkdf import HKDF



from .base import BaseBlockEngine, VerifyMessage, BCct



class BaseBCEllipticCurveEngine(BaseBlockEngine):
  MAX_ADDRESS_VALUE = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141
  
  
  def _get_pk(self, private_key : ec.EllipticCurvePrivateKey) -> ec.EllipticCurvePublicKey:
    """
    Simple wrapper to generate pk from sk


    Returns
    -------
    public_key : EllipticCurvePublicKey
    
    """
    return super(BaseBCEllipticCurveEngine, self)._get_pk(private_key)

  
  def _sk_to_text(
      self, 
      private_key :  ec.EllipticCurvePrivateKey, 
      password=None, 
      fn=None
    ):
    """
    Serialize a EllipticCurvePrivateKey as text

    Parameters
    ----------
    private_key : EllipticCurvePrivateKey
      the secret key object.
      
    password: str
      password to be used for sk serialization
      
    fn: str:
      text file where to save the pk

    Returns
    -------
      the sk as text string

    """
    return super(BaseBCEllipticCurveEngine, self)._sk_to_text(
      private_key=private_key, 
      password=password, 
      fn=fn
    ) 
  
  #############################################################################
  ##
  ##          MANDATORY DEFINITIONS:
  ##
  #############################################################################

  def _create_new_sk(self) -> ec.EllipticCurvePrivateKey:
    """
    Simple wrapper to generated pk


    Returns
    -------
    private_key : EllipticCurvePrivateKey
    
    """
    private_key = ec.generate_private_key(curve=ec.SECP256K1())
    return private_key

  def _create_new_sk_from_words(self, words: list[str]) -> ec.EllipticCurvePrivateKey:
      """
      Simple wrapper to generate pk using a seed

      Parameters
      ----------
      words : list[str]
          The words to be used as seed.

      Returns
      -------
      private_key : EllipticCurvePrivateKey
      """
      
      seedString = ';'.join(words)
      
      encodedString = seedString.encode()
      
      # Hash the seed to ensure it has enough entropy
      digest = hashlib.sha256(encodedString).digest()
      
      # Convert the hash to an integer
      private_value = int.from_bytes(digest, 'big')
      
      # Ensure the integer is within the valid range for the curve
      private_value = private_value % self.MAX_ADDRESS_VALUE

      # Create the private key from the integer
      private_key = ec.derive_private_key(private_value, ec.SECP256K1(), default_backend())
      return private_key
  
  def _sign(
      self, 
      data : bytes, 
      private_key : ec.EllipticCurvePrivateKey, 
      text=False
    ):
    """
    Sign a binary message with Elliptic Curve
    

    Parameters
    ----------
    data : bytes
      the binary message.
      
    private_key : ec.EllipticCurvePrivateKey
      the private key object.
      
    text : bool, optional
      return the signature as text. The default is False.

    Returns
    -------
    signature as text or binary

    """
    signature = private_key.sign(
      data=data,
      signature_algorithm=ec.ECDSA(hashes.SHA256())
      )
    txt_signature = self._binary_to_text(signature)
    return txt_signature if text else signature
    
  def _verify(
      self, 
      public_key : ec.EllipticCurvePublicKey, 
      signature : bytes, 
      data : bytes
    ):
    """
    Verifies a `EllipticCurvePublicKey` signature on a binary `data` package
    

    Parameters
    ----------
    public_key : ec.EllipticCurvePublicKey
      the pk object.
      
    signature : bytes
      the binary signature.
      
    data : bytes
      the binary message.


    Returns
    -------
    result: VerifyMessage 
      contains `result.valid` and `result.message`

    """
    result = VerifyMessage()
    try:
      public_key.verify(signature, data, ec.ECDSA(hashes.SHA256()))
      result.valid = True
      result.message = '<Signature OK>'
    except Exception as exp:
      err = str(exp)
      if len(err) == 0:
        err = exp.__class__.__name__
      result.message = err
      result.valid = False      
    return result    
  
  def _pk_to_address(self, public_key):
    """
    Given a EllipticCurvePublicKey object will return the simple text address


    Parameters
    ----------
    public_key : ec.EllipticCurvePublicKey
      the pk object.
      
    Returns
    -------
      text address      
    
    """
    data = public_key.public_bytes(
      encoding=serialization.Encoding.X962,
      format=serialization.PublicFormat.CompressedPoint,
    )
    txt = BCct.ADDR_PREFIX + self._binary_to_text(data)
    return txt


  def _address_to_pk(self, address):
    """
    Given a address will return the EllipticCurvePublicKey object


    Parameters
    ----------
    address : str
      the text address (pk).


    Returns
    -------
    pk : EllipticCurvePublicKey
      the pk object.

    """
    try:
      simple_address = self._remove_prefix(address)
      bpublic_key = self._text_to_binary(simple_address)
      public_key = ec.EllipticCurvePublicKey.from_encoded_point(
        curve=ec.SECP256K1(), 
        data=bpublic_key
      )
    except Exception as exp:
      self.P(f"Error converting address <{address}>to pk: {exp}", color='r')
      raise exp
    return public_key
  

  def __derive_shared_key(self, peer_public_key : str, info : str = BCct.DEFAULT_INFO, debug : bool = False):
    """
    Derives a shared key using own private key and peer's public key.

    Parameters
    ----------
    private_key : cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey
        The private key to use for derivation.
    peer_public_key : cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey
        The peer's public key.
    
    Returns
    -------
    bytes
        The derived shared key.
    """
    if info is not None:
      info = info.encode()
    private_key = self.private_key
    shared_key = private_key.exchange(ec.ECDH(), peer_public_key)
    if debug:
      print('sk-pk-shared_key: ', binascii.hexlify(shared_key).decode('utf-8'))
    derived_key = HKDF(
      algorithm=hashes.SHA256(),
      length=32,
      salt=None,
      info=info,
      backend=default_backend()
    ).derive(shared_key)
    if debug:
      print('derived-shared_key: ', base64.b64encode(derived_key))
    return derived_key


  def _encrypt(
    self, 
    plaintext: str, 
    receiver_address: str, 
    compressed: bool = True, # compressed is always True
    embed_compressed: bool = True, # embed_compressed is always True
    info: str = BCct.DEFAULT_INFO, 
    debug: bool = False,
  ):
    """
    Encrypts plaintext using the sender's private key and receiver's public key, 
    then base64 encodes the output.

    Parameters
    ----------
    receiver_address : str
        The receiver's address
        
    plaintext : str
        The plaintext to encrypt.
        
    compressed : bool, optional
        Whether to compress the plaintext before encryption. The default is True.
        
    embed_compressed : bool, optional
        Whether to embed the compressed flag in the encrypted data. The default is True.

    Returns
    -------
    str
        The base64 encoded nonce and ciphertext.
    """
    
    if compressed:
      to_encrypt_data = zlib.compress(plaintext.encode())
      compressed_flag = (1).to_bytes(1, byteorder='big')
    else:
      to_encrypt_data = plaintext.encode()
      compressed_flag = (0).to_bytes(1, byteorder='big')
      
    receiver_pk = self._address_to_pk(receiver_address)
    shared_key = self.__derive_shared_key(receiver_pk, info=info, debug=debug)
    aesgcm = AESGCM(shared_key)
    nonce = os.urandom(12)  # Generate a unique nonce for each encryption
    ciphertext = aesgcm.encrypt(nonce, to_encrypt_data, None)
    if embed_compressed:
      encrypted_data = nonce + compressed_flag + ciphertext
    else:
      encrypted_data = nonce + ciphertext  # Prepend the nonce to the ciphertext
    #end if      
    return base64.b64encode(encrypted_data).decode()  # Encode to base64
  
    
  
  def _decrypt(
    self, 
    encrypted_data_b64 : str, 
    sender_address : str, 
    decompress: bool = False, # decompress is only used if embed_compressed is False
    embed_compressed: bool = True,
    info: str = BCct.DEFAULT_INFO, 
    debug: bool = False
  ):
    """
    Decrypts base64 encoded encrypted data using the receiver's private key.
    
    The structure of the encrypted data is:
    - 12 bytes nonce
    - 1 byte compressed flag 
    - 13:... The ciphertext

    Parameters
    ----------        
    encrypted_data_b64 : str
        The base64 encoded nonce and ciphertext.
        
    sender_address : str
        The sender's address.
        
    decompress : bool, optional
        Whether to decompress the plaintext after decryption. The default is False as the decompression flag is embedded in the encrypted data.
        
    embed_compressed : bool, optional
        Whether the compressed flag is embedded in the encrypted data. The default is True.
        
        
    Returns
    -------
    str
        The decrypted plaintext.

    """
    try:
        
      sender_pk = self._address_to_pk(sender_address)
      encrypted_data = base64.b64decode(encrypted_data_b64)  # Decode from base64
      nonce = encrypted_data[:12]  # Extract the nonce      

      if embed_compressed:
        start_data = 13
        compressed_flag_byte = encrypted_data[12:13]
        compressed_flag = int.from_bytes(compressed_flag_byte, byteorder='big')
      else:
        start_data = 12
        compressed_flag = None      

      ciphertext = encrypted_data[start_data:]  # The rest is the ciphertext
      shared_key = self.__derive_shared_key(sender_pk, info=info, debug=debug)
      aesgcm = AESGCM(shared_key)
      plaintext = aesgcm.decrypt(nonce, ciphertext, None)
            
      if (embed_compressed and compressed_flag) or (not embed_compressed and decompress):
        plaintext = zlib.decompress(plaintext)          
      
      result = plaintext.decode()
    except Exception as exc:
      if debug:
        self.P("Error decrypting from <{}> (compressed_flag `{}`): {}".format(
          sender_address, compressed_flag, exc), color='r'
        )
      result = None
    return result
  

  ## Multi destination encryption
  def encrypt_for_multi(
    self, 
    plaintext: str, 
    receiver_addresses: list, 
    info: str = BCct.DEFAULT_INFO, 
    debug: bool = False
  ):
    """
    Encrypts plaintext for multiple receivers.
    
    The overall approach is to encrypt the plaintext with a symmetric key, 
    then encrypt the symmetric key with the public key of each receiver.

    Parameters
    ----------
    plaintext : str
        The plaintext to encrypt.
        
    receiver_addresses : list
        List of receiver addresses.

    Returns
    -------
    str
        The base64 encoded encrypted package.
    """
    to_encrypt_data = zlib.compress(plaintext.encode())
    compressed_flag = (1).to_bytes(1, byteorder='big')
    
    # Generate a random symmetric key
    symmetric_key = os.urandom(32)  # 256-bit key
    
    # Encrypt the plaintext with the symmetric key
    nonce = os.urandom(12)
    aesgcm = AESGCM(symmetric_key)
    ciphertext = aesgcm.encrypt(nonce, to_encrypt_data, None)
    
    # For each receiver, encrypt the symmetric key
    encrypted_keys = []
    for receiver_address in receiver_addresses:
      receiver_pk = self._address_to_pk(receiver_address)
      shared_key = self.__derive_shared_key(receiver_pk, info=info, debug=debug)
      # Use shared_key to encrypt the symmetric key
      aesgcm_shared = AESGCM(shared_key)
      nonce_shared = os.urandom(12)
      encrypted_symmetric_key = aesgcm_shared.encrypt(nonce_shared, symmetric_key, None)
      full_enc_key = nonce_shared + encrypted_symmetric_key
      encrypted_keys.append({
        'a': receiver_address,  # Address of the receiver
        'k': full_enc_key.hex() # Encrypted symmetric key
      })
    
    # Package the encrypted symmetric keys and the ciphertext
    encrypted_package = {
      'M': True,                  # Multi-recipient flag
      'c': compressed_flag.hex(), # Compressed flag
      'n': nonce.hex(),           # Nonce
      'd': ciphertext.hex(),      # Ciphertext
      'k': encrypted_keys         # Encrypted symmetric keys
    }
    
    # Convert to JSON, compress, and base64 encode
    enc_data = json.dumps(encrypted_package)
    enc_data_compressed = zlib.compress(enc_data.encode())
    enc_data_compressed_b64 = base64.b64encode(enc_data_compressed).decode()
    return enc_data_compressed_b64


  def decrypt_for_multi(
      self,
      encrypted_data_b64: str,
      sender_address: str,
      info: str = BCct.DEFAULT_INFO,
      debug: bool = False
  ):
    """
    Decrypts data encrypted for multiple receivers.

    Parameters
    ----------
    encrypted_data_b64 : str
      The base64 encoded encrypted package as produced by encrypt_for_multi.
    
    sender_address : str
      The sender's address (public key address) used to derive the shared key.
    
    info : str, optional
      Additional info used in the HKDF for shared key derivation.
    
    debug : bool, optional
      If True, prints debug information.

    Returns
    -------
    str or None
      The decrypted plaintext as a string, or None if decryption fails.
    """
    try:
      # 1. Base64 decode
      enc_data_compressed = base64.b64decode(encrypted_data_b64)

      # 2. Decompress the JSON structure
      enc_data_json = zlib.decompress(enc_data_compressed).decode()

      # 3. Parse JSON
      encrypted_package = json.loads(enc_data_json)

      # Expecting keys: 'M', 'c', 'n', 'd', 'k'
      # 'M' = True indicates multi-recipient
      if 'M' not in encrypted_package or encrypted_package['M'] != True:
        if debug:
          self.P("Not a multi-recipient package.", color='y')
        return None

      # 'c' = compressed flag (hex)
      compressed_flag_hex = encrypted_package['c']
      compressed_flag = int(compressed_flag_hex, 16)

      # 'n' = nonce (hex)
      nonce = bytes.fromhex(encrypted_package['n'])

      # 'd' = ciphertext (hex)
      ciphertext = bytes.fromhex(encrypted_package['d'])

      # 'k' = list of encrypted symmetric keys
      encrypted_keys = encrypted_package['k']

      # 4. Identify this receiver's encrypted symmetric key
      my_address = self.address
      my_encrypted_key_hex = None
      for ek in encrypted_keys:
        if ek['a'] == my_address:
          my_encrypted_key_hex = ek['k']
          break

      if my_encrypted_key_hex is None:
        if debug:
          self.P("No encrypted symmetric key for this receiver.", color='r')
        return None

      # Decode the encrypted symmetric key
      my_encrypted_key = bytes.fromhex(my_encrypted_key_hex)

      # The first 12 bytes are nonce_shared, rest is encrypted symmetric key
      nonce_shared = my_encrypted_key[:12]
      encrypted_symmetric_key = my_encrypted_key[12:]

      # 5. Derive shared key using sender's public key
      sender_pk = self._address_to_pk(sender_address)
      shared_key = self.__derive_shared_key(sender_pk, info=info, debug=debug)

      # 6. Decrypt the symmetric key
      aesgcm_shared = AESGCM(shared_key)
      symmetric_key = aesgcm_shared.decrypt(nonce_shared, encrypted_symmetric_key, None)

      # 7. Decrypt the ciphertext using the symmetric key
      aesgcm = AESGCM(symmetric_key)
      plaintext = aesgcm.decrypt(nonce, ciphertext, None)

      # 8. Decompress the plaintext if compressed_flag == 1
      if compressed_flag == 1:
        plaintext = zlib.decompress(plaintext)

      return plaintext.decode()

    except Exception as exc:
      if debug:
        self.P(f"Error decrypting multi scenario: {exc}", color='r')
      return None


  def encrypt(
    self, 
    plaintext: str, 
    receiver_address: any, 
    info: str = BCct.DEFAULT_INFO, 
    debug: bool = False,
    **kwargs
  ):
    """
    Encrypts plaintext using the sender's private key and receiver's public key, 
    then base64 encodes the output.

    Parameters
    ----------
    plaintext : str
        The plaintext to encrypt.

    receiver_address : str or list[str]
        The receiver's address or list of multiple receivers addresses.
        
        
    Obsolete:          
      compressed : bool, optional
          Whether to compress the plaintext before encryption. The default is True.
          
      embed_compressed : bool, optional
          Whether to embed the compressed flag in the encrypted data. The default is True.

    Returns
    -------
    str
        The base64 encoded nonce and ciphertext.
    """  
    assert isinstance(receiver_address, (str, list)), "receiver_address must be a string or a list of strings."
    if isinstance(receiver_address, list):
      return self.encrypt_for_multi(
        plaintext=plaintext, 
        receiver_addresses=receiver_address, 
        info=info, 
        debug=debug
      )      
    return self._encrypt(
      plaintext=plaintext, 
      receiver_address=receiver_address, 
      compressed=True, 
      embed_compressed=True, 
      info=info, 
      debug=debug,      
    )


  def decrypt(  
    self,
    encrypted_data_b64: str,
    sender_address: str,
    info: str = BCct.DEFAULT_INFO,
    debug: bool = False,
    is_multi: bool = True
  ):
    """
    Decrypts data encrypted for a single or multi  receiver.

    Parameters
    ----------
    encrypted_data_b64 : str
      The base64 encoded encrypted data.
    
    sender_address : str
      The sender's address (public key address) used to derive the shared key.
    
    info : str, optional
      Additional info used in the HKDF for shared key derivation.
    
    debug : bool, optional
      If True, prints debug information.
    
    is_multi : bool, optional
      If True, decrypts as multi-recipient package.

    Returns
    -------
    str or None
      The decrypted plaintext as a string, or None if decryption fails.
    """
    result = None
    if is_multi:
      result = self.decrypt_for_multi(encrypted_data_b64, sender_address, info=info, debug=debug)
    if result is None:
      result = self._decrypt(encrypted_data_b64, sender_address, decompress=True, embed_compressed=True, info=info, debug=debug)  
    return result
  
  
  ## end multi destination encryption
  




import json

from time import time
from copy import deepcopy

from ratio1 import Logger, const
from ratio1.bc import DefaultBlockEngine
from uuid import uuid4



if __name__ == '__main__' :
  l = Logger("ENC", base_folder='.', app_folder='_local_cache')
  eng = DefaultBlockEngine(
    log=l, name="default", 
    config={
        # "PEM_FILE": "aid01.pem",
      }
  )
  
  TEXT = "Please sign this message for Deeploy: "
  
  MSG_FROM_JSON = {"app_name":"app_40b23285","app_params":{"CR":"docker.io","CR_PASSWORD":"password","CR_USERNAME":"user","ENV":{"ENV1":"value1","ENV2":"value2","ENV3":"value3","ENV4":"value4"},"IMAGE":"repo/image:tag","OTHER_PARAM1":"value1","OTHER_PARAM2":"value2","OTHER_PARAM3":"value3","OTHER_PARAM4":"value4","OTHER_PARAM5":"value5","PORT":5000},"nonce":"0x195f1a86832","pipeline_input_type":"void","plugin_signature":"SOME_PLUGIN_01","target_nodes":["0xai_node_1","0xai_node_2"],"target_nodes_count":0,"EE_ETH_SIGN":"0x1048b65139affe77dffa59fbe7fd9586c0984fc73b235c7526140c87b2363b6609c0d5ed291ce6bb61756502129df704d5872dd7624ac59b7452a95dbe1a8fdf1b","EE_ETH_SENDER":"0xe558740ffc65bc73f6efb07c26c8d587ee22d297"}


  
  
  payload = deepcopy(MSG_FROM_JSON)
  sender = payload.get("EE_ETH_SENDER")

  addr = eng.eth_verify_payload_signature(
    payload=payload, message_prefix=TEXT,
    no_hash=True, indent=1,
  )
  valid = addr.upper() == sender.upper()
  if valid:
    l.P(f"Signature valid: {valid}, addr: {addr}")
  else:
    l.P(f"Signature invalid: {valid}, recovered: {addr} != {sender}")




import json


from ratio1 import Logger, const
from ratio1.bc import DefaultBlockEngine
from ratio1.utils.config import get_user_folder



if __name__ == '__main__' :
  l = Logger(
    "ENC", base_folder=str(get_user_folder()), 
    app_folder="_local_cache"
  )
  eng = DefaultBlockEngine(
    log=l, name="default", 
    config={        
      }
  )
  
  epochs = [
    1,2,3
  ]
  
  values = [
    10, 20, 30
  ]
  
  
  sign = eng.eth_sign_node_epochs(
    node=eng.eth_address, 
    epochs=epochs, 
    epochs_vals=values,
    signature_only=False,
  )
  
  l.P(f"Signature {sign}")


from ratio1 import Payload, Session, Pipeline, Instance


def instance_on_data(pipeline: Pipeline, data: Payload):
  # this could refresh a UI
  pipeline.P('Receive specific message from PERIMETER_VIOLATION_02:inst01 (hardcoded)', color='m')


def another_instance_on_data(pipeline: Pipeline, data: Payload):
  # this could refresh a UI
  pipeline.P('Receive specific message from CUSTOM_EXEC_01:inst01 (hardcoded)')


def pipeline_on_data(pipeline: Pipeline, signature, instance, data: Payload):
  # this could refresh a UI
  pipeline.P('Received data from box {} by server {}, stream:{}, plugin: {}, instance:{}, the following data:{}'.format(
      pipeline.node_id,  # can use pipeline.node_addr instead
      pipeline.session.server,
      pipeline.name,
      signature,
      instance,
      data
  ))


if __name__ == '__main__':

  node_id = 'stefan-box'

  sess = Session()

  pipeline: Pipeline = sess.create_pipeline(
      node=node_id,
      name='test_normal',
      data_source='VideoStream',
      config={
          'URL': 0
      },
      on_data=pipeline_on_data
  )

  # now we start a pet detector functionality for low-res cameras with all
  # the other params default
  pet_detector_instance: Instance = pipeline.create_plugin_instance(  # should return an id
      signature='PET_DETECTOR_01',
      instance_id='inst01',
      on_data=instance_on_data,
      ai_engine='lowres_general_detector'
  )

  plain_code = """
result="Data on node"
  """

  # now start a cyclic process
  custom_instance: Instance = pipeline.create_custom_plugin_instance(
      instance_id='inst01',
      plain_code=plain_code,
      on_data=another_instance_on_data
  )

  # now deploy the pipeline
  pipeline.deploy()

  # we wait for 30 seconds
  sess.run(30, close_session=False)

  # we can stop an instance like this
  pipeline.remove_plugin_instance(pet_detector_instance)

  # or like this
  custom_instance.stop()

  # we still need to deploy the changes
  pipeline.deploy()

  # we wait until the user presses `Ctrl+C`
  sess.run(0, close_session=False)

  # now close conn to comm server
  pipeline.close()
  sess.close()



import os
from ratio1 import Session


if __name__ == "__main__":
  AI4E_TUNNEL_ENGINE = "cloudflare"
  APP_NAME = "devnet-ai4e-api"

  target_node = os.getenv("AI4E_NODE")
  ai4e_token = os.getenv("AI4E_TOKEN")
  

  print(f"Using {AI4E_TUNNEL_ENGINE=} and {target_node=}", flush=True)

  assert target_node, "AI4E_NODE must be set in .env file"
  assert ai4e_token, "AI4E_TOKEN must be set in .env file"

  session = Session(
    evm_network='devnet'
  )
  

  pipeline, _ = session.create_web_app(
    name=APP_NAME,
    node=target_node,
    data_source="NetworkListener",
    signature="AI4EVERYONE",
    tunnel_engine=AI4E_TUNNEL_ENGINE,
    cloudflare_token=ai4e_token
  )
  
  pipeline.deploy()
  
  session.wait(
    seconds=10,
    close_session_on_timeout=True,
    close_pipeline_on_timeout=False
  )


class DistributedCustomCodePresets():
  PROCESS_REAL_TIME_COLLECTED_DATA__KEEP_UNIQUES_IN_AGGREGATED_COLLECTED_DATA = "PROCESS_REAL_TIME_COLLECTED_DATA__KEEP_UNIQUES_IN_AGGREGATED_COLLECTED_DATA"
  PROCESS_REAL_TIME_COLLECTED_DATA__KEEP_UNIQUES_IN_NODE_COLLECTED_DATA = "PROCESS_REAL_TIME_COLLECTED_DATA__KEEP_UNIQUES_IN_NODE_COLLECTED_DATA"
  PROCESS_REAL_TIME_COLLECTED_DATA__KEEP_IN_NODE_COLLECTED_DATA = "PROCESS_REAL_TIME_COLLECTED_DATA__KEEP_IN_NODE_COLLECTED_DATA"
  FINISH_CONDITION___EACH_NODE_COLLECTED_DATA_MORE_THAN_X = "FINISH_CONDITION___EACH_NODE_COLLECTED_DATA_MORE_THAN_X"
  FINISH_CONDITION___AGGREGATED_DATA_MORE_THAN_X = "FINISH_CONDITION___AGGREGATED_DATA_MORE_THAN_X"
  AGGREGATE_COLLECTED_DATA___AGGREGATE_COLLECTED_DATA = "AGGREGATE_COLLECTED_DATA___AGGREGATE_COLLECTED_DATA"
  AGGREGATE_COLLECTED_DATA___COLLECTED_DATA_PER_NODE = "AGGREGATE_COLLECTED_DATA___COLLECTED_DATA_PER_NODE"

  def __process_real_time_collected_data__keep_uniques_in_aggregated_collected_data(plugin, job_id, collected_data, data):
    all_data = sum(collected_data.values(), [])
    filtered_data = list(set(d for d in data if d not in all_data))
    return filtered_data

  def __process_real_time_collected_data__keep_uniques_in_node_collected_data(plugin, job_id, collected_data, data):
    filtered_data = list(set(d for d in data if d not in collected_data[job_id]))
    return filtered_data

  def __process_real_time_collected_data__keep_in_node_collected_data(plugin, job_id, collected_data, data):
    return data

  def __finish_condition___each_node_collected_data_more_than_x(plugin, collected_data):
    kwargs = plugin.cfg_finish_condition_kwargs
    if kwargs is None:
      raise Exception("`FINISH_CONDITION_KWARGS` must be set!")
    x = kwargs.get("X")
    if x is None:
      raise Exception("`X` must be set in `FINISH_CONDITION_KWARGS`!")
    return all([len(data) > x for data in collected_data])

  def __finish_condition___aggregated_data_more_than_x(plugin, collected_data):
    kwargs = plugin.cfg_finish_condition_kwargs
    if kwargs is None:
      raise Exception("`FINISH_CONDITION_KWARGS` must be set!")
    x = kwargs.get("X")
    if x is None:
      raise Exception("`X` must be set in `FINISH_CONDITION_KWARGS`!")
    return len(sum(collected_data.values(), [])) > x

  def __aggregate_collected_data___aggregate_collected_data(plugin, collected_data):
    return sum(collected_data.values(), [])

  def __aggregate_collected_data___collected_data_per_node(plugin, collected_data):
    return collected_data



from .base_logger import BaseLogger
from .logger_mixins import (_ClassInstanceMixin,
                            _ComputerVisionMixin,
                            _DateTimeMixin,
                            _DownloadMixin,
                            _GeneralSerializationMixin,
                            _JSONSerializationMixin,
                            _PickleSerializationMixin,
                            _ProcessMixin,
                            _ResourceSizeMixin,
                            _TimersMixin,
                            _UploadMixin,
                            _UtilsMixin,
                            _MachineMixin
                            )


class Logger(
  BaseLogger,
  _ClassInstanceMixin,
  _ComputerVisionMixin,
  _DateTimeMixin,
  _DownloadMixin,
  _GeneralSerializationMixin,
  _JSONSerializationMixin,
  _PickleSerializationMixin,
  _ProcessMixin,
  _ResourceSizeMixin,
  _TimersMixin,
  _UploadMixin,
  _UtilsMixin,
  _MachineMixin
):

  def __init__(self, lib_name="",
               lib_ver="",
               config_file="",
               config_data={},
               base_folder=None,
               app_folder=None,
               show_time=True,
               config_file_encoding=None,
               no_folders_no_save=False,
               max_lines=None,
               HTML=False,
               DEBUG=True,
               silent=False,
               data_config_subfolder=None,
               check_additional_configs=False,
               default_color='n',
               ):

    super(Logger, self).__init__(
      lib_name=lib_name, lib_ver=lib_ver,
      config_file=config_file,
      base_folder=base_folder,
      app_folder=app_folder,
      show_time=show_time,
      config_data=config_data,
      config_file_encoding=config_file_encoding,
      no_folders_no_save=no_folders_no_save,
      max_lines=max_lines,
      HTML=HTML,
      DEBUG=DEBUG,
      silent=silent,
      data_config_subfolder=data_config_subfolder,
      check_additional_configs=check_additional_configs,
      default_color=default_color,
    )
    self.cleanup_logs(archive_older_than_days=2)

    return

  def iP(self, str_msg, results=False, show_time=False, noprefix=False, color=None):
    if self.runs_from_ipython():
      return self._logger(
        str_msg,
        show=True, results=results, show_time=show_time,
        noprefix=noprefix, color=color
      )
    return

  def __repr__(self):
    # Get the name of the class
    class_name = self.__class__.__name__

    # Get public properties (those not starting with "_")
    public_properties = {k: v for k, v in self.__dict__.items() if not k.startswith("_")}

    # Convert properties to a string representation
    properties_str = ", ".join(f"{k}={v!r}" for k, v in public_properties.items())

    return f"{class_name}({properties_str})"


if __name__ == '__main__':
  l = Logger('TEST', base_folder='Dropbox', app_folder='_libraries_testdata')
  l.P("All check", color='green')
